import configparser
from datetime import date
import os

import audeer


config = configparser.ConfigParser()
config.read(os.path.join('..', 'setup.cfg'))

# Project -----------------------------------------------------------------
author = config['metadata']['author']
copyright = f'2021-{date.today().year} audEERING GmbH'
project = config['metadata']['name']
version = audeer.git_repo_version()
title = f'{project} Documentation'


# General -----------------------------------------------------------------
master_doc = 'index'
extensions = []
source_suffix = '.rst'
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', '**.ipynb_checkpoints']
pygments_style = None
extensions = [
    'jupyter_sphinx',  # executing code blocks
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',  # support for Google-style docstrings
    'sphinx.ext.viewcode',
    'sphinx.ext.intersphinx',
    'sphinx.ext.autosectionlabel',
    'sphinx_autodoc_typehints',
    'sphinx_copybutton',  # for "copy to clipboard" buttons
]

# Ignore package dependencies during building the docs
autodoc_mock_imports = [
    'tqdm',
]

# Reference with :ref:`data-header:Database`
autosectionlabel_prefix_document = True
autosectionlabel_maxdepth = 2

# Do not copy prompot output
copybutton_prompt_text = r'>>> |\.\.\. '
copybutton_prompt_is_regexp = True

# Mapping to external documentation
intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'torch': ('https://pytorch.org/docs/stable/', None),
    'audobject': ('https://audeering.github.io/audobject/', None),
}

# Disable Gitlab as we need to sign in
linkcheck_ignore = [
    'https://gitlab.audeering.com',
    'https://sphinx-doc.org',
]

# HTML --------------------------------------------------------------------
html_theme = 'sphinx_audeering_theme'
html_theme_options = {
    'display_version': True,
    'logo_only': False,
    'footer_links': False,
}
html_context = {
    'display_github': True,
}
html_title = title
html_static_path = ['_static']



import glob
import os
import random
import shutil

import numpy as np
import opensmile

import pytest
import torch

import audeer
import audiofile
import audobject


pytest.ROOT = audeer.safe_path(
    os.path.dirname(os.path.realpath(__file__))
)
pytest.TMP = audeer.mkdir(
    os.path.join(
        pytest.ROOT,
        audeer.uid(),
    )
)
pytest.SIGNAL, pytest.SAMPLING_RATE = audiofile.read(
    os.path.join(pytest.ROOT, 'test.wav'),
    always_2d=True,
)

# feature extractor

pytest.FEATURE = opensmile.Smile(
    feature_set=opensmile.FeatureSet.GeMAPSv01b,
    feature_level=opensmile.FeatureLevel.LowLevelDescriptors,
)
pytest.FEATURE_SHAPE = pytest.FEATURE(
    pytest.SIGNAL,
    pytest.SAMPLING_RATE,
).shape[1:]

# fix seed

seed = 1234
torch.backends.cudnn.deterministic = True
torch.manual_seed(seed)
torch.cuda.manual_seed_all(seed)
np.random.seed(seed)
random.seed(seed)


# create model with single output node

class TorchModelSingle(torch.nn.Module):

    def __init__(
        self,
    ):
        super().__init__()
        self.hidden = torch.nn.Linear(pytest.FEATURE_SHAPE[0], 8)
        self.out = torch.nn.Linear(8, 2)

    def forward(self, x: torch.Tensor):
        y = self.hidden(x.mean(dim=-1))
        y = self.out(y)
        return y.squeeze()


pytest.MODEL_SINGLE_PATH = os.path.join(pytest.TMP, 'single.onnx')
torch.onnx.export(
    TorchModelSingle(),
    torch.randn(pytest.FEATURE_SHAPE),
    pytest.MODEL_SINGLE_PATH,
    input_names=['feature'],
    output_names=['gender'],
    dynamic_axes={'feature': {1: 'time'}},
    opset_version=12,
)


# create model with multiple input and output nodes

class TorchModelMulti(torch.nn.Module):

    def __init__(
        self,
    ):

        super().__init__()

        self.hidden_left = torch.nn.Linear(1, 4)
        self.hidden_right = torch.nn.Linear(pytest.FEATURE_SHAPE[0], 4)
        self.out = torch.nn.ModuleDict(
            {
                'gender': torch.nn.Linear(8, 2),
                'confidence': torch.nn.Linear(8, 1),
            }
        )

    def forward(self, signal: torch.Tensor, spectrogram: torch.Tensor):

        y_left = self.hidden_left(signal.mean(dim=-1))
        y_right = self.hidden_right(spectrogram.mean(dim=-1))
        y_hidden = torch.cat([y_left, y_right], dim=-1)
        y_gender = self.out['gender'](y_hidden)
        y_confidence = self.out['confidence'](y_hidden)

        return (
            y_hidden.squeeze(),
            y_gender.squeeze(),
            y_confidence.squeeze(),
        )


pytest.MODEL_MULTI_PATH = os.path.join(pytest.TMP, 'multi.onnx')
torch.onnx.export(
    TorchModelMulti(),
    (
        torch.randn(pytest.SIGNAL.shape),
        torch.randn(pytest.FEATURE_SHAPE),
    ),
    pytest.MODEL_MULTI_PATH,
    input_names=['signal', 'feature'],
    output_names=['hidden', 'gender', 'confidence'],
    dynamic_axes={
        'signal': {1: 'time'},
        'feature': {1: 'time'},
    },
    opset_version=12,
)


# clean up

@pytest.fixture(scope='session', autouse=True)
def cleanup_session():
    path = os.path.join(
        pytest.TMP,
        '..',
        '.coverage.*',
    )
    for file in glob.glob(path):
        os.remove(file)
    yield
    if os.path.exists(pytest.TMP):
        shutil.rmtree(pytest.TMP)



from audonnx.core.model import (
    Model,
)
from audonnx.core.model import (
    InputNode,
    OutputNode,
)
from audonnx.core.api import (
    load,
)

__all__ = []


# Dynamically get the version of the installed module
try:
    import pkg_resources
    __version__ = pkg_resources.get_distribution(__name__).version
except Exception:  # pragma: no cover
    pkg_resources = None  # pragma: no cover
finally:
    del pkg_resources



import os
import typing

import numpy as np
import onnxruntime

import audeer
import audiofile
import audobject
import yaml

from audonnx.core.node import (
    InputNode,
    OutputNode,
)
from audonnx.core.typing import (
    Labels,
    Transform,
)


class Model(audobject.Object):
    r"""Model with multiple input and output nodes.

    For input nodes an optional transform can be given
    that transforms the raw signal into the desired representation,
    otherwise the raw signal is used as input.
    Use dictionary to assign transform objects to specific nodes
    if model has multiple input nodes.

    For output nodes an optional list of labels can be given,
    where each label corresponds to a dimension in the output,
    i.e. the number of labels must match the dimension of the output node.
    Use dictionary to assign labels to specific nodes
    if model has multiple output nodes.

    :class:`Model` inherites from :class:`audobject.Object`,
    which means you can seralize to
    and instantiate the class
    from a YAML file.
    Have a look at :class:`audobject.Object`
    to see all available methods.

    Args:
        path: path to model file
        labels: list of labels or dictionary with labels
        transform: callable object or a dictionary of callable objects

    """
    @audobject.init_decorator(
        borrow={
            'labels': '_original_args',
            'transform': '_original_args',
        },
        resolvers={
            'path': audobject.resolver.FilePath,
        },
    )
    def __init__(
            self,
            path: str,
            *,
            labels: Labels = None,
            transform: Transform = None,
    ):
        # keep original arguments to store them
        # when object is serialized
        self._original_args = {
            'labels': labels,
            'transform': transform,
        }

        self.path = audeer.safe_path(path)
        r"""Model path"""

        self.sess = onnxruntime.InferenceSession(self.path)
        r"""Interference session"""

        inputs = self.sess.get_inputs()
        outputs = self.sess.get_outputs()

        transform = transform or {}
        if not isinstance(transform, dict):
            if not len(inputs) == 1:
                names = [input.name for input in inputs]
                raise ValueError(
                    f'Model has multiple input nodes. '
                    f'Please use a dictionary to assign '
                    f'transform object to one of '
                    f'{names}.'
                )
            transform = {inputs[0].name: transform}

        labels = labels or {}
        if not isinstance(labels, dict):
            if not len(outputs) == 1:
                names = [output.name for output in outputs]
                raise ValueError(
                    f'Model has multiple output nodes. '
                    f'Please use a dictionary to assign '
                    f'labels to one of '
                    f'{names}.'
                )
            labels = {outputs[0].name: labels}

        self.inputs = {}
        r"""Input nodes"""
        for input in inputs:
            trans = transform[input.name] if input.name in transform else None
            self.inputs[str(input.name)] = InputNode(
                [-1 if x == 'time' else x for x in input.shape],
                input.type,
                trans,
            )

        self.outputs = {}
        r"""Output nodes"""
        for output in outputs:
            shape = output.shape or [1]
            dim = shape[-1]
            if output.name in labels:
                lab = labels[output.name]
                if len(lab) != dim:
                    raise ValueError(
                        f"Cannot assign "
                        f"{len(lab)} "
                        f"labels to output "
                        f"'{output.name}' "
                        f"with dimension "
                        f"{dim}."
                    )
            elif dim == 1:
                lab = [output.name]
            else:
                lab = [f'{output.name}-{idx}' for idx in range(dim)]
            self.outputs[output.name] = OutputNode(
                shape,
                output.type,
                lab,
            )

    def __call__(
            self,
            signal: np.ndarray,
            sampling_rate: int,
            *,
            output_names: typing.Union[str, typing.Sequence[str]] = None,
    ) -> typing.Union[
        np.ndarray,
        typing.Dict[str, np.ndarray],
    ]:
        r"""Compute output for one or more nodes.

        If ``output_names`` is a plain string,
        the output of the according node is returned.

        If ``output_names`` is a list of strings,
        a dictionary with according nodes as keys and
        their outputs as values is returned.

        If ``output_names`` is not set
        and the model has a single output node,
        the output of that node is returned.
        Otherwise a dictionary with outputs of all nodes is returned.

        Use :attr:`audonnx.Model.output_names` to get a list of available
        output nodes.

        Args:
            signal: input signal
            sampling_rate: sampling rate in Hz
            output_names: name of output or list with output names

        Returns:
            model output

        Examples:
            >>> audio_path = os.path.join('tests', 'test.wav')
            >>> signal, sampling_rate = audiofile.read(audio_path)
            >>> model_path = os.path.join('tests', 'model.yaml')
            >>> import audobject
            >>> model = audobject.from_yaml(model_path)
            >>> model(
            ...     signal,
            ...     sampling_rate,
            ...     output_names='gender',
            ... ).round(2)
            array([-195.1 ,   73.28], dtype=float32)

        """
        if output_names is None:
            output_names = list(self.outputs)
            if len(output_names) == 1:
                output_names = output_names[0]

        y = {}
        for name, input in self.inputs.items():
            if input.transform is not None:
                x = input.transform(signal, sampling_rate)
            else:
                x = signal
            y[name] = x.reshape(self.inputs[name].shape)

        z = self.sess.run(
            audeer.to_list(output_names),
            y,
        )

        if isinstance(output_names, str):
            z = z[0]
        else:
            z = {
                name: values for name, values in zip(output_names, z)
            }

        return z

    def __repr__(self) -> str:

        d = {
            'Input': {
                name: node._dict() for name, node in self.inputs.items()
            },
            'Output': {
                name: node._dict() for name, node in self.outputs.items()
            }
        }

        return yaml.dump(d, default_flow_style=None).strip()

    def __str__(self) -> str:
        return repr(self)

    def to_yaml(
            self,
            path: str,
            *,
            include_version: bool = True,
    ):
        r"""Save model to YAML file.

        Args:
            path: file path, must end on ``.yaml``
            include_version: add version to class name

        Raises:
            ValueError: if file path does not end on ``.yaml``

       """
        path = audeer.safe_path(path)
        if not audeer.file_extension(path) == 'yaml':
            raise ValueError(f"Model path {path} does not end on '.yaml'")
        with open(path, 'w') as fp:
            super().to_yaml(fp, include_version=include_version)

    # deprecated

    @audeer.deprecated(
        removal_version='0.4.0',
        alternative='__call__',
    )
    def forward(
            self,
            signal: np.ndarray,
            sampling_rate: int,
            *,
            output_names: typing.Union[str, typing.Sequence[str]] = None,
    ) -> typing.Union[
        np.ndarray,
        typing.Dict[str, np.ndarray],
    ]:  # pragma: no cover
        return self(signal, sampling_rate, output_names=output_names)

    @property
    @audeer.deprecated(
        removal_version='0.4.0',
        alternative='outputs[name].labels',
    )
    def labels(self) -> typing.Dict:  # pragma: no cover:
        return {
            name: output.labels for name, output in self.outputs.items()
        }

    @audeer.deprecated(
        removal_version='0.4.0',
    )
    def predict(
            self,
            signal: np.ndarray,
            sampling_rate: int,
            *,
            output_names: typing.Union[str, typing.Sequence[str]] = None,
    ) -> typing.Union[
        typing.Union[int, str],
        typing.Dict[str, typing.Union[int, str]],
    ]:  # pragma: no cover
        if output_names is None:
            output_names = list(self.outputs)
            if len(output_names) == 1:  # pragma: no cover
                output_names = output_names[0]

        y = self(
            signal,
            sampling_rate,
            output_names=audeer.to_list(output_names),
        )
        for name, values in y.items():
            y[name] = values.argmax()
            y[name] = self.labels[name][y[name]]

        if isinstance(output_names, str):
            y = y[output_names]

        return y

    @property
    @audeer.deprecated(
        removal_version='0.4.0',
        alternative='inputs[name].transform',
    )
    def transform(self) -> typing.Dict:  # pragma: no cover:
        return {
            name: input.transform for name, input in self.inputs.items()
        }



import numpy as np
import pytest

import audonnx


@pytest.mark.parametrize(
    'path, transform, labels, expected',
    [
        (
            pytest.MODEL_SINGLE_PATH,
            pytest.FEATURE,
            None,
            {'gender': ['gender-0', 'gender-1']},
        ),
        (
            pytest.MODEL_SINGLE_PATH,
            pytest.FEATURE,
            ['female', 'male'],
            {'gender': ['female', 'male']},
        ),
        pytest.param(
            pytest.MODEL_SINGLE_PATH,
            None,
            ['too', 'many', 'labels'],
            None,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
    ]
)
def test_labels(path, transform, labels, expected):
    model = audonnx.Model(
        path,
        labels=labels,
        transform=transform,
    )
    for name, l in expected.items():
        assert model.outputs[name].labels == l


@pytest.mark.parametrize(
    'model, output_names, expected',
    [
        (
            audonnx.Model(
                pytest.MODEL_SINGLE_PATH,
                transform=pytest.FEATURE,
            ),
            None,
            np.array([-195.1, 73.28], np.float32),
        ),
        (
            audonnx.Model(
                pytest.MODEL_SINGLE_PATH,
                transform=pytest.FEATURE,
            ),
            'gender',
            np.array([-195.1, 73.28], np.float32),
        ),
        (
            audonnx.Model(
                pytest.MODEL_SINGLE_PATH,
                transform=pytest.FEATURE,
            ),
            ['gender'],
            {'gender': np.array([-195.1, 73.28], np.float32)},
        ),
        (
            audonnx.Model(
                pytest.MODEL_MULTI_PATH,
                transform={
                    'feature': pytest.FEATURE,
                }
            ),
            None,
            {
                'hidden': np.array([
                    1.3299127e-01, 2.1280064e-01,
                    -4.7600174e-01, -3.7167081e-01,
                    6.6870685e+02, -4.2869656e+02,
                    -4.5552551e+02, 8.6153650e+02,
                ], np.float32),
                'gender': np.array([224.83, -12.72], np.float32),
                'confidence': np.array(-311.88, np.float32),
            },
        ),
        (
            audonnx.Model(
                pytest.MODEL_MULTI_PATH,
                transform={
                    'feature': pytest.FEATURE,
                }
            ),
            ['confidence', 'gender'],
            {
                'confidence': np.array(-311.88, np.float32),
                'gender': np.array([224.83, -12.72], np.float32),
            },
        ),
        (
            audonnx.Model(
                pytest.MODEL_MULTI_PATH,
                transform={
                    'feature': pytest.FEATURE,
                }
            ),
            'confidence',
            np.array(-311.88, np.float32),
        ),
    ],
)
def test_call(model, output_names, expected):
    y = model(
        pytest.SIGNAL,
        pytest.SAMPLING_RATE,
        output_names=output_names,
    )
    if isinstance(y, dict):
        for key, values in y.items():
            np.testing.assert_almost_equal(y[key], expected[key], decimal=2)
    else:
        np.testing.assert_almost_equal(y, expected, decimal=2)


@pytest.mark.parametrize(
    'path, labels, transform',
    [
        (
            pytest.MODEL_SINGLE_PATH,
            None,
            pytest.FEATURE,
        ),
        (
            pytest.MODEL_MULTI_PATH,
            {
                'gender': ['female', 'male'],
            },
            {
                'feature': pytest.FEATURE,
            }
        ),
        pytest.param(  # list of labels but multiple output nodes
            pytest.MODEL_MULTI_PATH,
            ['female', 'male'],
            None,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        pytest.param(  # transform object but multiple input nodes
            pytest.MODEL_MULTI_PATH,
            None,
            pytest.FEATURE,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
    ],
)
def test_nodes(path, labels, transform):

    model = audonnx.Model(
        path,
        labels=labels,
        transform=transform,
    )

    inputs = model.sess.get_inputs()
    outputs = model.sess.get_outputs()

    for idx, (name, node) in enumerate(model.inputs.items()):
        assert name == inputs[idx].name
        assert node.shape == [
            -1 if x == 'time' else x for x in inputs[idx].shape
        ]
        assert node.dtype == inputs[idx].type

    for idx, (name, node) in enumerate(model.outputs.items()):
        assert name == outputs[idx].name
        if outputs[idx].shape:
            assert node.shape == outputs[idx].shape
        else:
            assert node.shape == [1]
        assert node.dtype == outputs[idx].type


@pytest.mark.parametrize(
    'model, expected',
    [
        (
            audonnx.Model(
                pytest.MODEL_SINGLE_PATH,
                labels=['female', 'male'],
                transform=pytest.FEATURE,
            ), r'''Input:
  feature:
    shape: [18, -1]
    dtype: tensor(float)
    transform: opensmile.core.smile.Smile
Output:
  gender:
    shape: [2]
    dtype: tensor(float)
    labels: [female, male]'''
        ),
        (
            audonnx.Model(
                pytest.MODEL_MULTI_PATH,
                transform={
                    'feature': pytest.FEATURE,
                }
            ), r'''Input:
  signal:
    shape: [1, -1]
    dtype: tensor(float)
    transform: None
  feature:
    shape: [18, -1]
    dtype: tensor(float)
    transform: opensmile.core.smile.Smile
Output:
  hidden:
    shape: [8]
    dtype: tensor(float)
    labels: [hidden-0, hidden-1, hidden-2, (...), hidden-5, hidden-6, hidden-7]
  gender:
    shape: [2]
    dtype: tensor(float)
    labels: [gender-0, gender-1]
  confidence:
    shape: [1]
    dtype: tensor(float)
    labels: [confidence]'''  # noqa
        )
    ],
)
def test_str(model, expected):
    assert str(model) == expected



import os

import audeer
import numpy as np
import oyaml as yaml
import pytest

import audobject
import audonnx


@pytest.mark.parametrize(
    'path, transform, labels, expected',
    [
        (
            pytest.MODEL_SINGLE_PATH,
            pytest.FEATURE,
            {
                'gender': ['female', 'male'],
            },
            np.array([-195.1, 73.28], np.float32),
        ),
    ]
)
def test_load_legacy(tmpdir, path, transform, labels, expected):

    root = os.path.dirname(path)

    # create from onnx

    transform_path = os.path.join(root, 'transform.yaml')
    transform.to_yaml(transform_path)

    labels_path = os.path.join(root, 'labels.yaml')
    with open(labels_path, 'w') as fp:
        yaml.dump(labels, fp)

    model = audonnx.load(
        root,
        model_file=os.path.basename(path),
        transform_file=os.path.basename(transform_path),
        labels_file=os.path.basename(labels_path),
    )
    y = model(pytest.SIGNAL, pytest.SAMPLING_RATE)

    np.testing.assert_almost_equal(y, expected, decimal=2)
    for key, values in labels.items():
        assert model.outputs[key].labels == labels[key]

    # legacy mode -> load from ONNX if YAML does not exist

    model = audonnx.load(
        root,
        model_file=audeer.replace_file_extension(path, 'yaml'),
    )
    y = model(pytest.SIGNAL, pytest.SAMPLING_RATE)

    np.testing.assert_almost_equal(y, expected, decimal=2)
    for key, values in labels.items():
        assert model.outputs[key].labels == labels[key]

    # store wrong model YAML

    model_path = os.path.join(tmpdir, 'single.yaml')
    audobject.Object().to_yaml(model_path)
    model = audonnx.load(
        root,
        model_file=os.path.basename(path),
    )

    # file extension does not end on '.yaml'

    model_path = os.path.join(tmpdir, 'model.yml')
    with pytest.raises(ValueError):
        model.to_yaml(model_path)

    # create from YAML

    model_path = os.path.join(tmpdir, 'model.yaml')
    model.to_yaml(model_path)

    model = audonnx.load(tmpdir)
    y = model(pytest.SIGNAL, pytest.SAMPLING_RATE)

    np.testing.assert_almost_equal(y, expected, decimal=2)
    for key, values in labels.items():
        assert model.outputs[key].labels == labels[key]
